#include <cyg/kernel/kapi.h>
#include <cyg/hal/hal_tables.h>
#include <cyg/hal/plf_stub.h>
#include <cyg/io/io.h>
#include <net/etherio.h>
#include <network.h>
#include <cyg/infra/diag.h>

#include "../dm.h"
#include <cyg/hal/iic.h>
#include "../ident.h"

#define HW_ADDR_LEN 6

typedef struct
{
  unsigned char daddr[6];
  unsigned char saddr[6];
  unsigned short type;
} __attribute__((packed)) tEthHeader;

#define ETH_TYPE_COMTROL 0x11fe
#define ETH_HDR_SIZE     14
#define PRODUCT_ID       0xaa
#define ADMIN_CLASS      0x01
#define ASYNC_CLASS      0x55

#define ADMIN_ID_BOOT           0
#define ADMIN_BOOT_PACKET       1
#define ADMIN_ID_QUERY          2
#define ADMIN_ID_REPLY          3
#define ADMIN_ID_LOOP           4
#define ADMIN_ID_RESET          5
#define ADMIN_ID_TRACE          6
#define ADMIN_ID_INIT           7
#define ADMIN_ID_READMEM        8
#define ADMIN_ID_WRITEMEM       9
#define ADMIN_ID_TEST           10
#define ADMIN_ID_DATA           11
#define ADMIN_ID_ELBACK         12
#define ADMIN_ID_READIO         13
#define ADMIN_ID_WRITEIO        14
#define ADMIN_ID_RSMODE         15
#define ADMIN_ID_COMPMEM        16
#define ADMIN_ID_MOVEMEM        17
#define ADMIN_ID_TIMXMIT        18
 
#define ADMIN_IPINFO_SET    32
#define ADMIN_IPINFO_GET    33
#define ADMIN_IPINFO_ERASE  34
#define ADMIN_ID_MACADDR_SET    35
#define ADMIN_ID_MACADDR_GET    36
#define ADMIN_IPINFO_GETCUR 37
#define ADMIN_IPINFO_DISABLE 38

/* madmin task thread variables */
static unsigned madminStack[1024];
static cyg_handle_t  madminHandle;
static cyg_thread    madminThread;
static cyg_thread_entry_t madminTask;

/*
 *  Init function for the MAC mode admin thread.  This task implements Comtrol
 *  admin functions, that allow device control, administration and troubleshooting.
 */
void madminInit(void)
{
  diag_printf("starting madmin\n");
  cyg_thread_create(30, madminTask, (cyg_addrword_t)0, "madmin",
                    (void *)madminStack, sizeof madminStack, 
                    &madminHandle, &madminThread);
  cyg_thread_resume(madminHandle);
}

static void handleMacAdminCommand(tEthHeader *ep, unsigned char *req, int len);

static void done(void)
{
  diag_printf("madmin -- giving up!\n");
  while (1)
    cyg_thread_delay(100);
}

static unsigned char MyMacAddr[6];
cyg_io_handle_t EthHandle;

static void madminTask(cyg_addrword_t data)
{
  Cyg_ErrNo status;
  cyg_ether_info_t ethConfig;
  unsigned char *rxFrameBuf;
  tEthHeader *ehdr;
  tEthBuffer *rxBuffer;
  int len;
  
  // lookup and configure Ethernet port
  status = cyg_io_lookup("/dev/reth0", &EthHandle);

  if (status != ENOERR)
    {
      diag_printf("ERROR: cyg_io_lookup /dev/reth0 returned %d\n", status);
      done();
    }
  
  len = sizeof ethConfig;
  status = cyg_io_get_config(EthHandle, CYG_IO_GET_CONFIG_ETHER_INFO,&ethConfig,&len);
  if (status != ENOERR)
    {
      diag_printf("ERROR: cyg_io_get_config /dev/eth0 returned %d\n", status);
      done();
    }

  memcpy(MyMacAddr,ethConfig.mac_address,sizeof MyMacAddr);

#if (0)
  diag_printf("madmin MyMacAddr=%02x:%02x:%02x:%02x:%02x:%02x\n",
              MyMacAddr[0],
              MyMacAddr[1],
              MyMacAddr[2],
              MyMacAddr[3],
              MyMacAddr[4],
              MyMacAddr[5]);
#endif

  ethConfig.flags |= CYG_ETHER_FLAGS_WR_NONBLOCK;
  ethConfig.protocol = ETH_TYPE_COMTROL;

  len = sizeof ethConfig;
  status = cyg_io_set_config(EthHandle, CYG_IO_SET_CONFIG_ETHER_INFO,&ethConfig, &len);
  if (status != ENOERR)
    {
      diag_printf("ERROR: cyg_io_set_config returned %d\n",status);
      done();
    }

  while(1)
    {
      status = cyg_io_get_config(EthHandle, CYG_IO_GET_CONFIG_ETHER_RECV_BUFFER, &rxBuffer, NULL);
      if (status != ENOERR)
        {
          diag_printf("EtherRx: error %d\n", status);
          done();
        }
        
      rxFrameBuf = rxBuffer->data;
      len = rxBuffer->length;

#if 0
        {
          int i;
          diag_printf("ethrx [%d]: ",len);
          for (i=0; i<len; ++i)
            diag_printf("%02x ",rxFrameBuf[i]);
          diag_printf("\n");
        }
#endif
      ehdr = (tEthHeader*)&rxFrameBuf[2];

      if (ehdr->type == ETH_TYPE_COMTROL &&
          rxFrameBuf[16] == PRODUCT_ID &&
          rxFrameBuf[18] == ADMIN_CLASS )
        handleMacAdminCommand(ehdr, &rxFrameBuf[16], (len-(ETH_HDR_SIZE+2)));
      
      cyg_io_get_config(EthHandle,CYG_IO_GET_CONFIG_ETHER_FREE_BUFFER,rxBuffer, NULL);
    }
}


static int lanSendPkt(void *buf, int len)
{
  Cyg_ErrNo status;
  
  status = cyg_io_write(EthHandle, (unsigned char*)buf, &len);
  if (status != ENOERR)
    {
      diag_printf("LanSendPkt: Error %d\n",status);
      return -1;
    }

#if 0
    {
      int i;
      diag_printf("ethtx [%d]: ",len);
      for (i=0; i<len; ++i)
        diag_printf("%02x ",((unsigned char*)buf)[i]);
      diag_printf("\n");
    }
#endif
  return 0;
}


static void ReplyToIpinfoReq(unsigned char * reply_mac, unsigned char concIndex, unsigned char replyIndex,
                      int dataPresent, unsigned ipAddr, unsigned ipMask, unsigned ipGate)
{
  unsigned char *datap;
  tEthHeader *mac;
  unsigned char pkt[32];
      
  memset(pkt, 0, 32);
  mac = (tEthHeader *) pkt;
  memcpy(mac->daddr, reply_mac, HW_ADDR_LEN);
  memcpy(mac->saddr, ident.macAddr, HW_ADDR_LEN);
  mac->type = ETH_TYPE_COMTROL;
  datap = (unsigned char *) mac + ETH_HDR_SIZE;
  *datap++ = PRODUCT_ID;
  *datap++ = concIndex;
  *datap++ = ADMIN_CLASS;
  *datap++ = dataPresent ? 32 : 20;     /* length */
  *datap++ = 0;
  *datap++ = replyIndex;
  if (dataPresent)
    {
      *datap++ = ipAddr >> 24;
      *datap++ = ipAddr >> 16;
      *datap++ = ipAddr >> 8;
      *datap++ = ipAddr;
      *datap++ = ipMask >> 24;
      *datap++ = ipMask >> 16;
      *datap++ = ipMask >> 8;
      *datap++ = ipMask;
      *datap++ = ipGate >> 24;
      *datap++ = ipGate >> 16;
      *datap++ = ipGate >> 8;
      *datap++ = ipGate;
    }
  lanSendPkt(pkt, sizeof pkt);
  return;
}

static void ReplyToIdReq(unsigned char * reply_mac, unsigned char concIndex)
{
  unsigned char *data;
  tEthHeader *mac;
  unsigned char pkt[33 + MAX_VERSIONSTRING_SIZE + ETH_HDR_SIZE];
  unsigned long w;
  int lanPktLength, extDataLength = 0;
  
  memset(pkt, 0, sizeof pkt);

  mac = (tEthHeader*)pkt;
  if (reply_mac)
    memcpy(mac->daddr, reply_mac, HW_ADDR_LEN);
  else
    memset(mac->daddr, 0xff, HW_ADDR_LEN);
  memcpy(mac->saddr, ident.macAddr, HW_ADDR_LEN);
  mac->type = ETH_TYPE_COMTROL;
  
  data = pkt + ETH_HDR_SIZE;

  data[0] = PRODUCT_ID;
  data[1] = concIndex;
  data[2] = ADMIN_CLASS;
  data[3] = 27;                     // number of admin data bytes
  data[4] = 0;
  data[5] = ADMIN_ID_REPLY;

  // admin data starts here
  memcpy(&data[6], ident.macAddr, HW_ADDR_LEN);
  data[12] = 0x01;   // not owned
  data[13] = 1;                             // Bit 0 set, extended data present
  data[14] = 0;                             // reserved
  data[15] = 0;                             // reserved
  w = htonl(ident.modelId);
  memcpy(data+16,&w,4);                     // model ID
  data[20] = ident.archId;                  // architecture ID
  data[21] = ident.numPorts;                // number of ports
  data[22] = 0; data[23] = 2;               // program ID
  data[24] = ident.boardRev;                // board revision
  data[25] = 0;                             // boot rom major
  data[26] = 0;                             // boot rom minor
  data[27] = 1;                             // number of app records
  data[28] = 0; data[29] = 2;               // app id
  data[30] = ident.versionMajor;            // app version
  data[31] = ident.versionMinor;            // app version

  /*  Add the version string to the reply, if the global variable has a properly formed string. */
  if ((strlen(versionString) == 0) || (strlen(versionString) > MAX_VERSIONSTRING_SIZE - 1)) {
      diag_printf("Error: ReplyToIdReq() - Invalid version string length: %d \n",  strlen(versionString));
      data[32] = 0;     // Extended data type 1 - firmware ID string
      data[33] = 0;     // Extended data length
  }
  else {
      extDataLength = strlen(versionString) + 3;  // Extended data length (strlen + nullchar + type byte + len byte)

      data[32] = 1;                               // Extended data type 1 - firmware ID string
      data[33] = extDataLength;                   // Extended data length 
      strcpy(&data[34], versionString);           // The extended data
      data[32 + extDataLength] = 0;               // Terminate extended data
      data[32 + extDataLength + 1] = 0;
  }

  lanPktLength = 32 + ETH_HDR_SIZE + extDataLength + 2;
  lanSendPkt(pkt, sizeof pkt);
  return;
}


static void handleMacAdminCommand(tEthHeader *ehdr, unsigned char *data, int datalen)
{
  struct in_addr addr;
  unsigned long ipAddr, gateway, mask;

  if (data[5] == ADMIN_ID_RESET && data[6] == 0x55 && data[7] == 0x55)
    {
      diag_printf("Admin: reset\n");
      HAL_STUB_PLATFORM_RESET();
    }
  else if (data[5] == ADMIN_IPINFO_GETCUR)
    {
      diag_printf("Admin: Current IP info request recieved, response sent - ");
      ipAddr = addr.s_addr = myIpAddr();
      diag_printf("IP: %s ",inet_ntoa(addr));
      mask = addr.s_addr = myIpMask();
      diag_printf("Net Mask: %s ",inet_ntoa(addr));
      gateway = addr.s_addr = myIpGate();
      diag_printf("Gateway: %s \n",inet_ntoa(addr));
      ReplyToIpinfoReq(ehdr->saddr, data[1], ADMIN_IPINFO_GETCUR, 1,
                       ipAddr, mask, gateway);
      return;
    }
  else if (data[5] == ADMIN_IPINFO_GET)
    {
      unsigned long ipAddr = 0;
      unsigned long ipMask = 0;
      unsigned long ipGate = 0;
      int dataAvail;

      getIpInfoRt(&ipAddr,&ipMask,&ipGate);
      dataAvail = 1;
      diag_printf("Admin: Flash IP info request recieved, response sent - ");
      addr.s_addr = ipAddr;
      diag_printf("IP: %s ",inet_ntoa(addr));
      addr.s_addr = ipMask;
      diag_printf("Net Mask: %s ",inet_ntoa(addr));
      addr.s_addr = ipGate;
      diag_printf("Gateway: %s \n",inet_ntoa(addr));
      ReplyToIpinfoReq(ehdr->saddr, data[1], ADMIN_IPINFO_GET, dataAvail,
                       ipAddr, ipMask, ipGate);
      return;
    }
  else if (data[5] == ADMIN_IPINFO_GETCUR) // ??  Duplicate of above?
    {
      unsigned long ipAddr = 0;
      unsigned long ipMask = 0;
      unsigned long ipGate = 0;
      int dataAvail;
      dataAvail = 1;
      diag_printf("Admin: reporting current IP info %08x %08x %08x\n", ipAddr,
             ipMask, ipGate);
      ReplyToIpinfoReq(ehdr->saddr, data[1], ADMIN_IPINFO_GET, dataAvail,
                       ipAddr, ipMask, ipGate);
      return;
    }
  else if (data[5] == ADMIN_IPINFO_SET)
    {
      int dataAvail;
      unsigned long ipAddr = (data[6] << 24) + (data[7] << 16) + (data[8] << 8) + data[9];
      unsigned long ipMask = (data[10] << 24) + (data[11] << 16) + (data[12] << 8) + data[13];
      unsigned long ipGate = (data[14] << 24) + (data[15] << 16) + (data[16] << 8) + data[17];
      diag_printf("Admin: Flash IP set info request recieved, stored and made current - ");
      addr.s_addr = ipAddr;
      diag_printf("IP: %s ",inet_ntoa(addr));
      addr.s_addr = ipMask;
      diag_printf("Net Mask: %s ",inet_ntoa(addr));
      addr.s_addr = ipGate;
      diag_printf("Gateway: %s \n",inet_ntoa(addr));
      putIpInfoRt(ipAddr,ipMask,ipGate);
      ipAddr = ipMask = ipGate = 0;
      getIpInfoRt(&ipAddr,&ipMask,&ipGate);
      dataAvail = 1;
      ReplyToIpinfoReq(ehdr->saddr, data[1], ADMIN_IPINFO_SET, 1, 
                       ipAddr,ipMask, ipGate);
      return;
    }
  else if (data[5] == ADMIN_IPINFO_ERASE || data[5] == ADMIN_IPINFO_DISABLE)
    {
      unsigned char *datap, pkt[32];
      unsigned char cmd = data[5];
      tEthHeader *mac;
      
      memset(pkt, 0, sizeof pkt);
      mac = (tEthHeader *) pkt;
      memcpy(mac->daddr, ehdr->saddr, HW_ADDR_LEN);
      memcpy(mac->saddr, ident.macAddr, HW_ADDR_LEN);
      mac->type = ETH_TYPE_COMTROL;
      datap = (unsigned char *) mac + ETH_HDR_SIZE;
      *datap++ = PRODUCT_ID;
      *datap++ = data[1];
      *datap++ = ADMIN_CLASS;
      *datap++ = 1;             /* length */
      *datap++ = 0;
      *datap++ = cmd;
      lanSendPkt(mac, 20);
      if (cmd == ADMIN_IPINFO_DISABLE)
        {
          diag_printf("Admin: IP Disable\n");
          putIpInfoRt(0xffffffff,0,0);
        }
      else
        {
          diag_printf("Admin: IP Erase\n");
          putIpInfoRt(0,0,0);
        }
    }
  else if (data[5] == ADMIN_ID_QUERY)
    {
      unsigned char *replymac = data+6;
      if (((replymac[0] == 0) && (replymac[1] == 0) &&
           (replymac[2] == 0) && (replymac[3] == 0) &&
           (replymac[4] == 0) && (replymac[5] == 0)) ||
          ((replymac[0] == 0xff) && (replymac[1] == 0xff) &&
           (replymac[2] == 0xff) && (replymac[3] == 0xff) &&
           (replymac[4] == 0xff) && (replymac[5] == 0xff)))
        replymac = ehdr->saddr;
      ReplyToIdReq(replymac, data[1]);
    }
}
