/*******************************************************************
 *       commands.c
 *
 *       Common command parsing functions for telnet servers.
 *******************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <cyg/kernel/kapi.h>
#include <cyg/hal/hal_tables.h>
#include <cyg/hal/plf_stub.h>
#include <cyg/infra/diag.h>
#include <cyg/hal/iic.h>
#include "dm.h"
#include "commands.h"
#include "string.h"
#include "ident.h"
#include "state.h"

/*  Global externs */
extern char versionString[64];
extern char linkTimeStamp[];
extern tServerState serverState[];

/*  Globals */
const char prompt[] = "\r\n\rdm> ";

/*  Statics Variables */
static const char prodName[] = "Comtrol DeviceMaster RTS ";
static char responseString[MAX_RESPONSE_STRING];
static char introPrompt[MAX_RESPONSE_STRING];
static char ipBytes[4];
static char ipString[80];

/*  Statics Function Prototypes */
static char *resetCommand(char **tokens, int numberofArgs);
static char *verCommand(char **tokens, int numberofArgs);
static char *ipCommand(char **tokens, int numberofArgs);
static char *helpCommand(char **tokens, int numberofArs);
static char *getAddrBytes(unsigned long addr);
static char *getIpString(void); 
static char *getMacString(void);
static char *macCommand(char **tokens, int numberofArgs);
static char *passwordCommand(char **tokens, int numberofArgs);
static char *timeoutCommand(char **tokens, int numberofArgs);
static char *telnetCommand(char **tokens, int numberofArgs);
static char *teltimeoutCommand(char **tokens, int numberofArgs);
static char *modelCommand(char **tokens, int numberofArgs);
static char *setbaudCommand(char **tokens, int numberofArgs);
static unsigned long getIpFromOctets(char *str);

/*  The command array */
termCommands_t termCommands[] = {
    {"reset",      resetCommand,      0, "reset      - Resets the device"},
    {"ip",         ipCommand,         3, "ip         - View/set IP address"},
    {"timeout",    timeoutCommand,    1, "timeout    - Set time (in seconds) until Socket Server loads automatically"},
    {"mac",        macCommand,        0, "mac        - View MAC address"},
    {"password",   passwordCommand,   1, "password   - Set the password"},
    {"telnet",     telnetCommand,     1, "telnet     - Enable/disable telnet"},
    {"teltimeout", teltimeoutCommand, 1, "teltimeout - Set the telnet timeout period (in seconds)"},
    {"model",      modelCommand,      0, "model      - View the Model ID"},

#ifndef NO_SOCKETSERVER
    {"setbaud",    setbaudCommand,    3, "setbaud    - Set the baud on all ports"},
#endif

    {"ver",        verCommand,        0, "ver        - Display firmware revision"},
    {"help",       helpCommand,       1, "help       - Display this help info"}
};

/*  Resets the device */
static char *resetCommand(char **tokens, int numberofArgs)
{
    cyg_thread_delay(10);
    HAL_STUB_PLATFORM_RESET();
    return "Reset Function";
}

/*  Resets the device */
static char *verCommand(char **tokens, int numberofArgs)
{
    return GetIntroPrompt();
}

/*  Views the Model ID */
static char *modelCommand(char **tokens, int numberofArgs)
{
    sprintf(responseString, "Model ID:  %d \n\r", 
	    (unsigned int)ident.modelId);
    return responseString;
}

/*  Sets the timeout period, the time before Socket Server starts automatically. */
static char *timeoutCommand(char **tokens, int numberofArgs)
{
    unsigned char timeout;

    if (numberofArgs == 1) {
	timeout = atol(tokens[1]);
	if (timeout >= 255)
	    ident.timeoutValue = 255;
	else
	    ident.timeoutValue = timeout;
	putRedbootDelayRt(ident.timeoutValue);
    }

    sprintf(responseString, "Timeout %u seconds \n\r", 
	    (unsigned int)ident.timeoutValue);
    return responseString;
}

/*  Sets the telnet timeout period.  Value must be > 30.  0=infinite.  */
static char *teltimeoutCommand(char **tokens, int numberofArgs)
{
    unsigned long timeout;

    if (numberofArgs == 1) {
	timeout =atol(tokens[1]);
	if ((timeout > 0) && (timeout < 30))
	    ident.telnetTimeout = 30;
	else
	    ident.telnetTimeout = atol(tokens[1]);
	putTelnetTimeoutRt(ident.telnetTimeout);
    }
    sprintf(responseString, "Telnet timeout %u seconds \n\r", 
	    (unsigned int)ident.telnetTimeout);
    return responseString;
}

/*  Enables/disables telnet access */
static char *telnetCommand(char **tokens, int numberofArgs)
{
    int originalSetting;

    originalSetting = ident.telnetEnable;
    memset(responseString, 0, sizeof(responseString));
    if (numberofArgs == 1) {
	if (!strcmp(tokens[1], "enable")) {
	    ident.telnetEnable = 1;
	    putTelnetEnableRt(ident.telnetEnable);
	}
	else if (!strcmp(tokens[1], "disable")) {
	    ident.telnetEnable = 0;
	    putTelnetEnableRt(ident.telnetEnable);
	}
	else {
	    strcpy(responseString, "  Invalid argument  ");
	}
    }
    if (ident.telnetEnable)
	strcat(responseString, "Telnet is enabled");
    else
	strcat(responseString, "Telnet is disabled");
    if (originalSetting != ident.telnetEnable)
	strcat(responseString, ".  Reset device for change to take effect. ");
    return responseString;
}


/*  Sets the password */
static char *passwordCommand(char **tokens, int numberofArgs)
{
    if (numberofArgs == 0) {
	ident.password[0] = 0x00;
	strcpy(responseString, "Password cleared");
    }
    else {
	strcpy(ident.password, tokens[1]);
	strcpy(responseString, "Password set");
    }
    putPasswordRt(ident.password);
    return responseString;
}

/* Generates the standard help screen */
static void generateHelp(void)
{
    char str[80];
    int i;

    memset(responseString, 0, sizeof(responseString));
    sprintf(str, "\n\r");
    strcat(responseString, str);
    for (i = 0; i < sizeof(termCommands) / sizeof(termCommands_t); i++) {
	sprintf(str, "%s \n\r", termCommands[i].help);
	strcat(responseString, str);
    }
    sprintf(str, "quit       - Exit session \n\r");
    strcat(responseString, str);
}

/*  Display command help screen */
static char *helpCommand(char **tokens, int numberofArgs)
{
    char str[80];

    if (numberofArgs == 0) {
	generateHelp();
    }
    else if (numberofArgs == 1) {
	if (!strcmp(tokens[1],"ip")) {
	    sprintf(str, "\n\r  Setting the IP Address \n\r");
	    strcpy(responseString, str);
	    sprintf(str, "ip  -  Displays the present IP/Mask/Gateway \n\r");
	    strcat(responseString, str);
	    sprintf(str, "ip <address>  -  Ex.  ip 192.168.250.250, uses default mask and gateway \n\r");
	    strcat(responseString, str);
	    sprintf(str, "ip <address> <mask>   -  Ex.  ip 192.168.250.250 255.255.0.0, uses default gateway \n\r");
	    strcat(responseString, str);
	    sprintf(str, "ip <address> <mask> <gateway> -  Ex.  ip 192.168.250.250 255.255.0.0 192.168.0.1 \n\r");
	    strcat(responseString, str);
	    sprintf(str, "NOTE:  Reset required to store and use IP address change \n\r\n\r");
	    strcat(responseString, str);
	}
#if (0)
	else if (!strcmp(tokens[1],"mac")) {
	    sprintf(str, "\n\r  Setting the MAC Address \n\r");
	    strcpy(responseString, str);
	    sprintf(str, "mac  -  Displays the present MAC Address \n\r");
	    strcat(responseString, str);
	    sprintf(str, "mac <address>  -  Ex.  ip 00 C0 4E 12 00 FF \n\r");
	    strcat(responseString, str);
	    sprintf(str, "NOTE:  Reset required to store and use MAC address change \n\r\n\r");
	    strcat(responseString, str);
	}
#endif
	else if (!strcmp(tokens[1],"password")) {
	    sprintf(str, "\n\r  Sets the RTS password \n\r");
	    strcpy(responseString, str);
	    sprintf(str, "password  -  Clears the password (there will be no password) \n\r");
	    strcat(responseString, str);
	    sprintf(str, "password <pwd> - Sets the password to <pwd> \n\r");
	    strcat(responseString, str);
	}
	else if (!strcmp(tokens[1],"telnet")) {
	    sprintf(str, "\n\r  Enable or disable telnet access \n\r");
	    strcpy(responseString, str);
	    sprintf(str, "telnet  -  Displays the present setting \n\r");
	    strcat(responseString, str);
	    sprintf(str, "telnet enable - Enables telnet access \n\r");
	    strcat(responseString, str);
	    sprintf(str, "telnet disable - Disables telnet access \n\r\n\r");
	    strcat(responseString, str);
	}
	else if (!strcmp(tokens[1],"teltimeout")) {
	    sprintf(str, "\n\r  Sets the telnet timeout value (in seconds) \n\r");
	    strcpy(responseString, str);
	    sprintf(str, "teltimeout  -  Displays the present setting \n\r");
	    strcat(responseString, str);
	    sprintf(str, "teltimeout <value> - Ex.  teltimeout 300, sets timeout to 300 seconds \n\r");
	    strcat(responseString, str);
	    sprintf(str, "teltimeout 0 - Sets timeout infinite. \n\r");
	    strcat(responseString, str);
	    sprintf(str, "NOTE:  Telnet timeout is required since DeviceMaster telnet is single session only. \n\r");
	    strcat(responseString, str);
	    sprintf(str, "       Only one user can be connected at a time.  Infinite timeout is discouraged. \n\r\n\r");
	    strcat(responseString, str);
	}
	else if (!strcmp(tokens[1],"timeout")) {
	    sprintf(str, "\n\r  Sets the time (in seconds) before Socket Server loads automatically \n\r");
	    strcpy(responseString, str);
	    sprintf(str, "timeout  -  Displays the present setting \n\r");
	    strcat(responseString, str);
	    sprintf(str, "timeout <value> - Ex.  teltimeout 60, sets timeout to 60 seconds \n\r");
	    strcat(responseString, str);
	    sprintf(str, "timeout 0 - Socket Server never loads. \n\r");
	    strcat(responseString, str);
	}
	else {
	    generateHelp();
	}
    }
	
    return responseString;
}

/*  Returns the standard introductory prompt */
char *GetIntroPrompt(void)
{
    char aLine[80], *aLinePtr;

    /*  Display product, version, timestamp of build */
    sprintf(aLine, "\n\r%s  ModelID: %u \n\r\r",prodName, (unsigned int)ident.modelId);
    strcpy(introPrompt, aLine);
    sprintf(aLine, "%s \n\r",versionString);
    strcat(introPrompt, aLine);
    sprintf(aLine, "Built: %s \n\r",linkTimeStamp);
    strcat(introPrompt, aLine);

    /*  Display ip/mask/gateway */
    aLinePtr = getIpString();
    strcat(introPrompt, aLinePtr);

     /*  Display MAC */
    aLinePtr = getMacString();
    strcat(introPrompt, aLinePtr);

    sprintf(aLine, "\n\r");
    strcat(introPrompt, aLine);
    return introPrompt;
}

/*
 *  Processed commands recieved from the client.
 */
char *processIncomingCmd(char *incoming_command)
{
    char cmd[MAX_SIZEOF_COMMAND], *tokens[NUMBEROF_TOKENS], delim = ' ';
    int i, numberofTokens = 0;

    memcpy(cmd, incoming_command, MAX_SIZEOF_COMMAND);

    for (i = 0; i < NUMBEROF_TOKENS; i++) {
	if (i == 0) 
	    tokens[i] = strtok(cmd, &delim);
	else 
	    tokens[i] = strtok(NULL, &delim);
	if (tokens[i] == NULL)
	    break;
	numberofTokens++;
    }

    for (i = 0; i < sizeof(termCommands) / sizeof(termCommands_t); i++) {
	if (!strcmp(tokens[0], termCommands[i].theCommand)) {
	    if ((numberofTokens - 1) > termCommands[i].numberofArgs) 
		return "Invalid number of arguments";
	    return termCommands[i].fp(tokens, numberofTokens - 1);
	}
    }
    return "Invalid Command";
}

/*  Processes IP commands */
static char *ipCommand(char **tokens, int numberofArgs)
{
    if (numberofArgs == 0) {
	return getIpString();
    } 

    if (numberofArgs == 1) {
	ident.ipAddr = getIpFromOctets(tokens[1]);
	ident.netMask = 0xFFFF0000;                       // 255.255.0.0
	ident.gateway = (ident.ipAddr & 0xFFFFFF00) | 1;  // 192.168.0.1
    }
    else if (numberofArgs == 2) {
	ident.ipAddr = getIpFromOctets(tokens[1]);
	ident.netMask = getIpFromOctets(tokens[2]);
	ident.gateway = (ident.ipAddr & 0xFFFFFF00) | 1;  
    }
    else {
	ident.ipAddr = getIpFromOctets(tokens[1]);
	ident.netMask = getIpFromOctets(tokens[2]);
	ident.gateway = getIpFromOctets(tokens[3]);
    }
    putIpInfoRt(ident.ipAddr, ident.netMask, ident.gateway);
    strcpy(responseString, getIpString());
    strcat(responseString, "IP Address stored, reset to take effect. ");
    return responseString;
}

#ifndef NO_SOCKETSERVER

/*  Processes setbaud commands */
static char *setbaudCommand(char **tokens, int numberofArgs)
{
    unsigned long BaudTable[] = 
	{300, 1200, 2400, 9600, 14400, 19200, 28800, 38400, 57600, 115200, 230400, 0};
    static char usage[160] = "Sets and enables/disables Socket Server baudrate on all ports.  \n\rUsage: setbaud <baud> [enable/disable] [rtscts]";
    static char baudUsage[160] = "Invalid baud - 300, 1200, 2400, 9600, 14400, 19200, 28800, 38400, 57600, 115200, 230400";
    unsigned long baud;
    int i = 0, rtscts = 0, enable = 1;

    if ((numberofArgs < 1) ||  (numberofArgs > 3)) 
	return usage;

    //  Get and verify the baud rate
    baud =atol(tokens[1]);
    while (1) {
	if (BaudTable[i] == 0)
	    return baudUsage;
	if (baud == BaudTable[i])
	    break;
	i++;
    }

    if (numberofArgs == 2) 
	if (!strcmp(tokens[2], "disable"))
	    enable = 0;

    if (numberofArgs == 3) 
	if (!strcmp(tokens[3], "rtscts"))
	    rtscts = 1;
	
    for (i = 0; i < ident.numPorts; i++) {
	if (enable)
	    serverState[i].config.enabled = 1;
	else
	    serverState[i].config.enabled = 0;
	serverState[i].config.listenEnable = 1;
	serverState[i].config.listenPort = 8000 + i;
	serverState[i].config.connectToIP = 0;
	serverState[i].config.connectToPort = 0;
	serverState[i].config.connectAlways = 0;
	serverState[i].config.connectOnCD = 0;
	serverState[i].config.connectOnDSR = 0;
	serverState[i].config.disconnOnIdle = 0;
	serverState[i].config.disconnOnNoCD = 0;
	serverState[i].config.disconnOnNoDSR = 0;
	serverState[i].config.idleTimeout = 300;
	serverState[i].config.interfaceMode = Mode232;
	if (rtscts)
	    serverState[i].config.flowControl = FlowRTSCTS;
	else
	    serverState[i].config.flowControl = FlowNone;
	serverState[i].config.parity = ParityNone;
	serverState[i].config.dtrControl = DTRoff;
	serverState[i].config.baudRate = baud;
	serverState[i].config.wordLength = 8;
	serverState[i].config.stopBits = 1;
    }
    serverSaveConfig();

    sprintf(responseString, "Socket Server baudrate set on all ports to %d ", (int)baud);
    if (rtscts)
	strcat(responseString, "RTS-CTS enabled");
    return responseString;
}

#endif


/*  Creates 4 bytes from an IP address */
static char *getAddrBytes(unsigned long addr)
{
    ipBytes[0] = (addr & 0xFF000000) >> 24;
    ipBytes[1] = (addr & 0x00FF0000) >> 16;
    ipBytes[2] = (addr & 0x0000FF00) >> 8;
    ipBytes[3] = (addr & 0x000000FF);
    return ipBytes;
}

/*  Display ip/mask/gateway */
static char *getIpString(void) 
{
    char aLine[80], *octets;

    octets = getAddrBytes(ident.ipAddr);
    sprintf(aLine, "IP Addr: %d.%d.%d.%d ",
	    octets[0], octets[1], octets[2], octets[3]);
    strcpy(ipString, aLine);

    octets = getAddrBytes(ident.netMask);
    sprintf(aLine, " Mask: %d.%d.%d.%d ",
	    octets[0], octets[1], octets[2], octets[3]);
    strcat(ipString, aLine);

    octets = getAddrBytes(ident.gateway);
    sprintf(aLine, "Gateway: %d.%d.%d.%d \n\r",
	    octets[0], octets[1], octets[2], octets[3]);
    strcat(ipString, aLine);

    return ipString;
}

static unsigned long getIpFromOctets(char *str)
{
    char ipString[80], octets[4], delim = '.', *octetString;
    int i;

    strcpy(ipString, str);
    memset(octets, 0, sizeof(octets));
    for (i = 0; i < 4; i++ ) {
	if (i == 0) 
	    octetString = strtok(ipString, &delim);
	else 
	    octetString = strtok(NULL, &delim);
	if (octetString == NULL)
	    break;
	else
	    octets[i] = (char)(atoi(octetString));
    }

    return ((octets[0] << 24) |
	    (octets[1] << 16) |
	    (octets[2] << 8) |
	    (octets[3]));
}

/*  Process MAC commands */
/*  NOTE:  It was decided not to allows MAC changes from telnet */
static char *macCommand(char **tokens, int numberofArgs)
{
#if (0)
    int i;

    if (numberofArgs == 0) {
	strcpy(responseString, getMacString());
    }
    else if (numberofArgs == 6) {
	for (i = 0; i < 6; i++ ) 
	    ident.macAddr[i] = atoi(tokens[i + 1]);
	putMacAddress(ident.macAddr);
	strcpy(responseString, getMacString());
	strcat(responseString, "MAC Address stored, reset to take effect. ");
    }
    else {
	strcpy(responseString, "Invalid number of arguments, see help mac");
    }
    return responseString;
#endif
    return getMacString();
}

/*  Display MAC */
static char *getMacString(void) 
{
    sprintf(responseString, "MAC Addr: %02x %02x %02x %02x %02x %02x \n\r",
	    ident.macAddr[0], ident.macAddr[1], ident.macAddr[2], 
	    ident.macAddr[3], ident.macAddr[4], ident.macAddr[5]);

    return responseString;
}

