#include <cyg/kernel/kapi.h>
#include <cyg/hal/hal_tables.h>
#include <cyg/hal/plf_stub.h>
#include <cyg/infra/diag.h>
#include <network.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

#include "../dm.h"
#include <cyg/hal/iic.h>
#include "../ident.h"
#include "../commands.h"

// state for background diagnostic task
static unsigned telnetsrvStack[2048];
static cyg_handle_t  telnetsrvHandle;
static cyg_thread    telnetsrvThread;
static cyg_thread_entry_t telnetsrvTask;

/*
 *  Init function for the telnet server task.  This allows a minimum of 
 *  system configuration through telnet.
 */
void telnetsrvInit(void)
{
  if (!ident.telnetEnable)
    return;
  diag_printf("starting telnetsrv\n");
  cyg_thread_create(30, telnetsrvTask, (cyg_addrword_t)0, "telnetsrv",
                    (void *)telnetsrvStack, sizeof telnetsrvStack, 
                    &telnetsrvHandle, &telnetsrvThread);
  cyg_thread_resume(telnetsrvHandle);
}

#define TELNET_SB   250
#define TELNET_SE   240
#define TELNET_WILL 251
#define TELNET_WONT 252
#define TELNET_DO   253
#define TELNET_DONT 254
#define TELNET_IAC  255

static void telnetReply(int sock, unsigned char resp)
{
  unsigned char msg[3];
  msg[0] = TELNET_IAC;
  msg[1] = resp;
  read(sock,&msg[2],1);
  write(sock,msg,sizeof msg);
}

static int telnetGets(int sock, char *p, int max)
{
  char *start = p;
  unsigned char c;
  static unsigned char last;
  fd_set rfds;
  struct timeval tv;
  int retval;
  
  FD_ZERO(&rfds);
  FD_SET(sock, &rfds);
  tv.tv_sec = ident.telnetTimeout;
  tv.tv_usec =0;
 
  while (1)
    {
	if (tv.tv_sec == 0)
	    retval = select(sock +1, &rfds, NULL, NULL, NULL);
	else
	    retval = select(sock +1, &rfds, NULL, NULL, &tv);
      if( !retval )
        return -2;

      if (read(sock,&c,1) <= 0) 
        return -1;
      
      if (c=='\n' && last =='\r')
        continue;

      last = c;
      if (c=='\n' || c=='\r')
        {
          *p = '\0';
          return (p - start);  
        }
      
      if (c == TELNET_IAC)
        {
          if (read(sock,&c,1) <= 0) return -1;
          switch (c)
            {
             case TELNET_WILL: telnetReply(sock,TELNET_DONT); break;
             case TELNET_DO:   telnetReply(sock,TELNET_WONT); break;
             case TELNET_WONT: telnetReply(sock,TELNET_DONT); break;
             case TELNET_DONT: telnetReply(sock,TELNET_WONT); break;
             case TELNET_SB:
              do
                {
                  if (read(sock,&c,1) <= 0)
                    return -1;
                }
              while (c != TELNET_SE);
              break;
             default:
              // ignore it, I guess
              break;
            }
        }
      else if (p < (start+max) - 1)
        *p++ = c;
    }
}

static const char pwprompt[] = "\r\nPassword: ";
static const char pwinvalid[] = "\r\nPassword invalid\r\n";

static char line[128];
static void handleCommand(int sock, char *cmdLine);
static int terseMode;

static void telnetsrvTask(cyg_addrword_t data)
{
  int lsock;
  int on = 1,rc;
  struct sockaddr_in slocal,sremote;
  extern int TcpStackOk;
  char *intro;

  diag_printf("telnetsrv running\n");
  
  while (!TcpStackOk)
    cyg_thread_delay(50);
  
  lsock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
  if (lsock < 0) 
    {
      diag_printf("Cannot create socket.\n");
      return;
    }
  
  memset(&slocal, 0, sizeof(slocal));
  slocal.sin_family = AF_INET;
  slocal.sin_len = sizeof(slocal);
  slocal.sin_port = 23;
  slocal.sin_addr.s_addr = INADDR_ANY;
 
  rc = bind(lsock, (struct sockaddr *)&slocal, sizeof(slocal));
  if (rc) 
    {
      diag_printf("Cannot bind socket: %d %d\n",rc,errno);
    }
  
  if (setsockopt(lsock, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on))) 
    diag_printf("Cannot set ReuseAddr\n");
  if (setsockopt(lsock, SOL_SOCKET, SO_REUSEPORT, &on, sizeof(on))) 
    diag_printf("Cannot set ReusePort\n");
  
listen(lsock, 1);
  
  while (1)
    {
      int dsock;
      int len = sizeof(sremote);
      dsock = accept(lsock, (struct sockaddr *)&sremote, &len);
      if (dsock < 0)
        {
	  diag_printf("telnetsrv: failure on accept\n");
        }
      diag_printf("telnetsrv: connection from %s:%d\n", inet_ntoa(sremote.sin_addr),sremote.sin_port);

      // prompt for password
      write(dsock,pwprompt,strlen(pwprompt));
      
      if ((telnetGets(dsock, line, sizeof line) < 0)
          || strcmp(ident.password,line))
        {
          write(dsock,pwinvalid,strlen(pwinvalid));
          diag_printf("telnetserv: invalid password\n");
          cyg_thread_delay(200);
          close(dsock);
          continue;
        }
      
      intro = GetIntroPrompt();
      write(dsock, intro, strlen(intro));

      while (1)
        {

          if (!terseMode)
	      write(dsock,prompt, strlen(prompt));
          if (telnetGets(dsock,line,sizeof line) < 0)
            break;
          handleCommand(dsock,line);
        }
      diag_printf("telnetserv: closed\n");
      close(dsock);
    }
}

static void handleCommand(int dsock, char *cmdLine)
{
    char *responseString;

    /*  Check for "quit" */
    if ((cmdLine[0] == 'q')  || (!strcmp(cmdLine, "quit"))) { 
      close(dsock);
      return;
    }

    responseString = processIncomingCmd(cmdLine);
    write(dsock, responseString, strlen(responseString));
}

//=============================================================================================================
